@@ -68,24 +68,31 @@ struct group_commit_orderer {
   */
   bool installed;
 
-  /*
-    This flag is set for a GCO in which we have event groups with multiple
-    different commit_id values from the master. This happens when we
-    optimistically try to execute in parallel transactions not known to be
-    conflict-free.
+  enum force_switch_bits
+  {
+    /*
+      This flag is set for a GCO in which we have event groups with multiple
+      different commit_id values from the master. This happens when we
+      optimistically try to execute in parallel transactions not known to be
+      conflict-free.
 
-    When this flag is set, in case of DDL we need to start a new GCO regardless
-    of current commit_id, as DDL is not safe to speculatively apply in parallel
-    with prior event groups.
-  */
-  static const uint8 MULTI_BATCH = 0x01;
-  /*
-    This flag is set for a GCO that contains DDL. If set, it forces a switch to
-    a new GCO upon seeing a new commit_id, as DDL is not safe to speculatively
-    replicate in parallel with subsequent transactions.
-  */
-  static const uint8 FORCE_SWITCH = 0x02;
+      When this flag is set, in case of DDL we need to start a new GCO
+      regardless of current commit_id, as DDL is not safe to
+      speculatively apply in parallel with prior event groups.
+    */
+    MULTI_BATCH= 1,
+    /*
+      This flag is set for a GCO that contains DDL. If set, it forces
+      a switch to a new GCO upon seeing a new commit_id, as DDL is not
+      safe to speculatively replicate in parallel with subsequent
+      transactions.
+    */
+    FORCE_SWITCH= 2
+  };
   uint8 flags;
+#ifndef DBUG_OFF
+  bool gc_done;
+#endif
 };
 
 
@@ -240,6 +247,8 @@ struct rpl_parallel_thread_pool {
   rpl_parallel_thread_pool();
   int init(uint32 size);
   void destroy();
+  void deactivate();
+  void destroy_cond_mutex();
   struct rpl_parallel_thread *get_thread(rpl_parallel_thread **owner,
                                          rpl_parallel_entry *entry);
   void release_thread(rpl_parallel_thread *rpt);
@@ -341,7 +350,8 @@ struct rpl_parallel_entry {
   group_commit_orderer *current_gco;
 
   rpl_parallel_thread * choose_thread(rpl_group_info *rgi, bool *did_enter_cond,
-                                      PSI_stage_info *old_stage, bool reuse);
+                                      PSI_stage_info *old_stage,
+                                      Gtid_log_event *gtid_ev);
   int queue_master_restart(rpl_group_info *rgi,
                            Format_description_log_event *fdev);
 };
