@@ -124,8 +124,7 @@ public:
   /** Create temporary sp_name object from MDL key. Store in qname_buff */
   sp_name(const MDL_key *key, char *qname_buff);
 
-  ~sp_name()
-  {}
+  ~sp_name() = default;
 };
 
 
@@ -625,6 +624,23 @@ public:
     DBUG_RETURN(false);
   }
 
+  /**
+    Iterate through the LEX stack from the top (the newest) to the bottom
+    (the oldest) and find the one that contains a non-zero spname.
+    @returns - the address of spname, or NULL of no spname found.
+  */
+  const sp_name *find_spname_recursive()
+  {
+    uint count= m_lex.elements;
+    for (uint i= 0; i < count; i++)
+    {
+      const LEX *tmp= m_lex.elem(count - i - 1);
+      if (tmp->spname)
+        return tmp->spname;
+    }
+    return NULL;
+  }
+
   /// Put the instruction on the backpatch list, associated with the label.
   int
   push_backpatch(THD *thd, sp_instr *, sp_label *);
@@ -1286,8 +1302,7 @@ public:
     m_query.length= 0;
   }
 
-  virtual ~sp_instr_stmt()
-  {};
+  virtual ~sp_instr_stmt() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1322,8 +1337,7 @@ public:
       m_lex_keeper(lex, lex_resp)
   {}
 
-  virtual ~sp_instr_set()
-  {}
+  virtual ~sp_instr_set() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1366,8 +1380,7 @@ public:
       m_field_offset(field_offset)
   {}
 
-  virtual ~sp_instr_set_row_field()
-  {}
+  virtual ~sp_instr_set_row_field() = default;
 
   virtual int exec_core(THD *thd, uint *nextp);
 
@@ -1409,8 +1422,7 @@ public:
       m_field_name(field_name)
   {}
 
-  virtual ~sp_instr_set_row_field_by_name()
-  {}
+  virtual ~sp_instr_set_row_field_by_name() = default;
 
   virtual int exec_core(THD *thd, uint *nextp);
 
@@ -1436,8 +1448,7 @@ public:
       value(val), m_lex_keeper(lex, TRUE)
   {}
 
-  virtual ~sp_instr_set_trigger_field()
-  {}
+  virtual ~sp_instr_set_trigger_field() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1480,8 +1491,7 @@ public:
       m_dest(dest), m_cont_dest(0), m_optdest(0), m_cont_optdest(0)
   {}
 
-  virtual ~sp_instr_opt_meta()
-  {}
+  virtual ~sp_instr_opt_meta() = default;
 
   virtual void set_destination(uint old_dest, uint new_dest)
     = 0;
@@ -1510,8 +1520,7 @@ public:
     : sp_instr_opt_meta(ip, ctx, dest)
   {}
 
-  virtual ~sp_instr_jump()
-  {}
+  virtual ~sp_instr_jump() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1562,8 +1571,7 @@ public:
       m_lex_keeper(lex, TRUE)
   {}
 
-  virtual ~sp_instr_jump_if_not()
-  {}
+  virtual ~sp_instr_jump_if_not() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1610,8 +1618,7 @@ public:
     : sp_instr(ip, ctx)
   {}
 
-  virtual ~sp_instr_preturn()
-  {}
+  virtual ~sp_instr_preturn() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1642,8 +1649,7 @@ public:
       m_lex_keeper(lex, TRUE)
   {}
 
-  virtual ~sp_instr_freturn()
-  {}
+  virtual ~sp_instr_freturn() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1748,8 +1754,7 @@ public:
     : sp_instr(ip, ctx), m_count(count)
   {}
 
-  virtual ~sp_instr_hpop()
-  {}
+  virtual ~sp_instr_hpop() = default;
 
   void update_count(uint count)
   {
@@ -1782,8 +1787,7 @@ public:
     m_frame(ctx->current_var_count())
   {}
 
-  virtual ~sp_instr_hreturn()
-  {}
+  virtual ~sp_instr_hreturn() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1819,8 +1823,7 @@ public:
     : sp_instr(ip, ctx), m_lex_keeper(lex, TRUE), m_cursor(offset)
   {}
 
-  virtual ~sp_instr_cpush()
-  {}
+  virtual ~sp_instr_cpush() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1854,8 +1857,7 @@ public:
     : sp_instr(ip, ctx), m_count(count)
   {}
 
-  virtual ~sp_instr_cpop()
-  {}
+  virtual ~sp_instr_cpop() = default;
 
   void update_count(uint count)
   {
@@ -1887,8 +1889,7 @@ public:
     : sp_instr(ip, ctx), m_cursor(c)
   {}
 
-  virtual ~sp_instr_copen()
-  {}
+  virtual ~sp_instr_copen() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1925,8 +1926,7 @@ public:
       m_cursor(coffs),
       m_var(voffs)
   {}
-  virtual ~sp_instr_cursor_copy_struct()
-  {}
+  virtual ~sp_instr_cursor_copy_struct() = default;
   virtual int execute(THD *thd, uint *nextp);
   virtual int exec_core(THD *thd, uint *nextp);
   virtual void print(String *str);
@@ -1948,8 +1948,7 @@ public:
     : sp_instr(ip, ctx), m_cursor(c)
   {}
 
-  virtual ~sp_instr_cclose()
-  {}
+  virtual ~sp_instr_cclose() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -1978,8 +1977,7 @@ public:
     m_varlist.empty();
   }
 
-  virtual ~sp_instr_cfetch()
-  {}
+  virtual ~sp_instr_cfetch() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -2017,8 +2015,7 @@ public:
   sp_instr_agg_cfetch(uint ip, sp_pcontext *ctx)
     : sp_instr(ip, ctx){}
 
-  virtual ~sp_instr_agg_cfetch()
-  {}
+  virtual ~sp_instr_agg_cfetch() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -2043,8 +2040,7 @@ public:
     : sp_instr(ip, ctx), m_errcode(errcode)
   {}
 
-  virtual ~sp_instr_error()
-  {}
+  virtual ~sp_instr_error() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
@@ -2077,8 +2073,7 @@ public:
       m_lex_keeper(lex, TRUE)
   {}
 
-  virtual ~sp_instr_set_case_expr()
-  {}
+  virtual ~sp_instr_set_case_expr() = default;
 
   virtual int execute(THD *thd, uint *nextp);
 
