@@ -1,4 +1,5 @@
-/* Copyright 2008-2020 Codership Oy <http://www.codership.com>
+/* Copyright 2008-2022 Codership Oy <http://www.codership.com>
+   Copyright (c) 2008, 2022, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,9 +30,11 @@
 #include "wsrep_utils.h"
 #include "wsrep_xid.h"
 #include "wsrep_thd.h"
+#include "wsrep_server_state.h"
 
 #include <cstdio>
 #include <cstdlib>
+#include "debug_sync.h"
 
 #include <my_service_manager.h>
 
@@ -265,6 +268,12 @@ static bool sst_auth_real_set (const char* value)
       if (wsrep_sst_auth) { my_free((void*) wsrep_sst_auth); }
       wsrep_sst_auth= my_strdup(PSI_INSTRUMENT_ME, WSREP_SST_AUTH_MASK, MYF(0));
     }
+    else
+    {
+      if (wsrep_sst_auth) { my_free((void*) wsrep_sst_auth); }
+      wsrep_sst_auth= NULL;
+    }
+
     return 0;
   }
   return 1;
@@ -308,12 +317,45 @@ bool wsrep_before_SE()
 }
 
 // Signal end of SST
-static void wsrep_sst_complete (THD*                thd,
-                                int const           rcode)
+static bool wsrep_sst_complete (THD*                thd,
+                                int const           rcode,
+                                wsrep::gtid const   sst_gtid)
 {
   Wsrep_client_service client_service(thd, thd->wsrep_cs());
-  Wsrep_server_state::instance().sst_received(client_service, rcode);
+  Wsrep_server_state& server_state= Wsrep_server_state::instance();
+  enum wsrep::server_state::state state= server_state.state();
+  bool failed= false;
+  char start_pos_buf[FN_REFLEN];
+  ssize_t len= wsrep::print_to_c_str(sst_gtid, start_pos_buf, FN_REFLEN-1);
+  start_pos_buf[len]='\0';
+
+  // Do not call sst_received if we are not in joiner or
+  // initialized state on server. This is because it
+  // assumes we are on those states. Give error if we are
+  // in incorrect state.
+  if ((state == Wsrep_server_state::s_joiner ||
+       state == Wsrep_server_state::s_initialized))
+  {
+    if (Wsrep_server_state::instance().sst_received(client_service, rcode))
+    {
+      failed= true;
+    }
+    else
+    {
+      WSREP_INFO("SST succeeded for position %s", start_pos_buf);
+    }
+  }
+  else
+  {
+    WSREP_ERROR("SST failed for position %s initialized %d server_state %s",
+                start_pos_buf,
+                server_state.is_initialized(),
+                wsrep::to_c_string(state));
+    failed= true;
+  }
+
   wsrep_joiner_monitor_end();
+  return failed;
 }
 
   /*
@@ -325,13 +367,15 @@ static void wsrep_sst_complete (THD*                thd,
   @param seqno     [IN]               Initial state sequence number
   @param state     [IN]               Always NULL, also ignored by wsrep provider (?)
   @param state_len [IN]               Always 0, also ignored by wsrep provider (?)
+  @return true when successful, false if error
 */
-void wsrep_sst_received (THD*                thd,
+bool wsrep_sst_received (THD*                thd,
                          const wsrep_uuid_t& uuid,
                          wsrep_seqno_t const seqno,
                          const void* const   state,
                          size_t const        state_len)
 {
+  bool error= false;
   /*
     To keep track of whether the local uuid:seqno should be updated. Also, note
     that local state (uuid:seqno) is updated/checkpointed only after we get an
@@ -371,8 +415,10 @@ void wsrep_sst_received (THD*                thd,
     if (WSREP_ON)
     {
       int const rcode(seqno < 0 ? seqno : 0);
-      wsrep_sst_complete(thd,rcode);
+      error= wsrep_sst_complete(thd,rcode, sst_gtid);
     }
+
+    return error;
 }
 
 static int sst_scan_uuid_seqno (const char* str,
@@ -596,6 +642,7 @@ static void* sst_joiner_thread (void* a)
             goto err;
           } else {
             wsrep_gtid_server.domain_id= (uint32) domain_id;
+            wsrep_gtid_domain_id= (uint32)domain_id;
           }
         }
       }
@@ -653,7 +700,7 @@ err:
     /* Read committed isolation to avoid gap locking */
     thd->variables.tx_isolation= ISO_READ_COMMITTED;
 
-    wsrep_sst_complete (thd, -err);
+    wsrep_sst_complete (thd, -err, ret_gtid);
 
     delete thd;
     my_thread_end();
@@ -662,51 +709,75 @@ err:
   return NULL;
 }
 
-#define WSREP_SST_AUTH_ENV "WSREP_SST_OPT_AUTH"
+#define WSREP_SST_AUTH_ENV        "WSREP_SST_OPT_AUTH"
+#define WSREP_SST_REMOTE_AUTH_ENV "WSREP_SST_OPT_REMOTE_AUTH"
+#define DATA_HOME_DIR_ENV         "INNODB_DATA_HOME_DIR"
 
-static int sst_append_auth_env(wsp::env& env, const char* sst_auth)
+static int sst_append_env_var(wsp::env&   env,
+                              const char* const var,
+                              const char* const val)
 {
-  int const sst_auth_size= strlen(WSREP_SST_AUTH_ENV) + 1 /* = */
-    + (sst_auth ? strlen(sst_auth) : 0) + 1 /* \0 */;
+  int const env_str_size= strlen(var) + 1 /* = */
+                          + (val ? strlen(val) : 0) + 1 /* \0 */;
 
-  wsp::string sst_auth_str(sst_auth_size); // for automatic cleanup on return
-  if (!sst_auth_str()) return -ENOMEM;
+  wsp::string env_str(env_str_size); // for automatic cleanup on return
+  if (!env_str()) return -ENOMEM;
 
-  int ret= snprintf(sst_auth_str(), sst_auth_size, "%s=%s",
-                    WSREP_SST_AUTH_ENV, sst_auth ? sst_auth : "");
+  int ret= snprintf(env_str(), env_str_size, "%s=%s", var, val ? val : "");
 
-  if (ret < 0 || ret >= sst_auth_size)
+  if (ret < 0 || ret >= env_str_size)
   {
-    WSREP_ERROR("sst_append_auth_env(): snprintf() failed: %d", ret);
+    WSREP_ERROR("sst_append_env_var(): snprintf(%s=%s) failed: %d",
+                var, val, ret);
     return (ret < 0 ? ret : -EMSGSIZE);
   }
 
-  env.append(sst_auth_str());
+  env.append(env_str());
   return -env.error();
 }
 
-#define DATA_HOME_DIR_ENV "INNODB_DATA_HOME_DIR"
-
-static int sst_append_data_dir(wsp::env& env, const char* data_dir)
-{
-  int const data_dir_size= strlen(DATA_HOME_DIR_ENV) + 1 /* = */
-    + (data_dir ? strlen(data_dir) : 0) + 1 /* \0 */;
-
-  wsp::string data_dir_str(data_dir_size); // for automatic cleanup on return
-  if (!data_dir_str()) return -ENOMEM;
-
-  int ret= snprintf(data_dir_str(), data_dir_size, "%s=%s",
-                    DATA_HOME_DIR_ENV, data_dir ? data_dir : "");
-
-  if (ret < 0 || ret >= data_dir_size)
-  {
-    WSREP_ERROR("sst_append_data_dir(): snprintf() failed: %d", ret);
-    return (ret < 0 ? ret : -EMSGSIZE);
-  }
-
-  env.append(data_dir_str());
-  return -env.error();
-}
+#ifdef __WIN__
+/*
+  Space, single quote, ampersand, backquote, I/O redirection
+  characters, caret, all brackets, plus, exclamation and comma
+  characters require text to be enclosed in double quotes:
+*/
+#define IS_SPECIAL(c) \
+  (isspace(c) || c == '\'' || c == '&' || c == '`' || c == '|' || \
+                 c ==  '>' || c == '<' || c == ';' || c == '^' || \
+                 c ==  '[' || c == ']' || c == '{' || c == '}' || \
+                 c ==  '(' || c == ')' || c == '+' || c == '!' || \
+                 c ==  ',')
+/*
+  Inside values, equals character are interpreted as special
+  character and requires quotation:
+*/
+#define IS_SPECIAL_V(c) (IS_SPECIAL(c) || c == '=')
+/*
+  Double quotation mark and percent characters require escaping:
+*/
+#define IS_REQ_ESCAPING(c) (c == '""' || c == '%')
+#else
+/*
+  Space, single quote, ampersand, backquote, and I/O redirection
+  characters require text to be enclosed in double quotes. The
+  semicolon is used to separate shell commands, so it must be
+  enclosed in double quotes as well:
+*/
+#define IS_SPECIAL(c) \
+  (isspace(c) || c == '\'' || c == '&' || c == '`' || c == '|' || \
+                 c ==  '>' || c == '<' || c == ';')
+/*
+  Inside values, characters are interpreted as in parameter names:
+*/
+#define IS_SPECIAL_V(c) IS_SPECIAL(c)
+/*
+  Double quotation mark and backslash characters require
+  backslash prefixing, the dollar symbol is used to substitute
+  a variable value, therefore it also requires escaping:
+*/
+#define IS_REQ_ESCAPING(c) (c == '"' || c == '\\' || c == '$')
+#endif
 
 static size_t estimate_cmd_len (bool* extra_args)
 {
@@ -732,11 +803,17 @@ static size_t estimate_cmd_len (bool* extra_args)
       char c;
       while ((c = *arg++) != 0)
       {
-        /* A whitespace or a single quote requires double quotation marks: */
-        if (isspace(c) || c == '\'')
+        if (IS_SPECIAL(c))
         {
           quotation= true;
         }
+        else if (IS_REQ_ESCAPING(c))
+        {
+          cmd_len++;
+#ifdef __WIN__
+          quotation= true;
+#endif
+        }
         /*
           If the equals symbol is encountered, then we need to separately
           process the right side:
@@ -755,49 +832,20 @@ static size_t estimate_cmd_len (bool* extra_args)
           }
           while ((c = *arg++) != 0)
           {
-            /*
-              A whitespace or a single quote requires double
-              quotation marks:
-            */
-            if (isspace(c) || c == '\'')
+            if (IS_SPECIAL_V(c))
             {
               quotation= true;
             }
-            /*
-              Double quotation mark or backslash symbol requires backslash
-              prefixing:
-            */
-#ifdef __WIN__
-            else if (c == '"' || c == '\\')
-#else
-            /*
-              The dollar symbol is used to substitute a variable, therefore
-              it also requires escaping:
-            */
-            else if (c == '"' || c == '\\' || c == '$')
-#endif
+            else if (IS_REQ_ESCAPING(c))
             {
               cmd_len++;
+#ifdef __WIN__
+              quotation= true;
+#endif
             }
           }
           break;
         }
-        /*
-          Double quotation mark or backslash symbol requires backslash
-          prefixing:
-        */
-#ifdef __WIN__
-        else if (c == '"' || c == '\\')
-#else
-        /*
-          The dollar symbol is used to substitute a variable, therefore
-          it also requires escaping:
-        */
-        else if (c == '"' || c == '\\' || c == '$')
-#endif
-        {
-          cmd_len++;
-        }
       }
       /* Perhaps we need to quote the entire argument or its right part: */
       if (quotation)
@@ -840,11 +888,17 @@ static void copy_orig_argv (char* cmd_str)
       char c;
       while ((c = *arg_scan++) != 0)
       {
-        /* A whitespace or a single quote requires double quotation marks: */
-        if (isspace(c) || c == '\'')
+        if (IS_SPECIAL(c))
         {
           quotation= true;
         }
+        else if (IS_REQ_ESCAPING(c))
+        {
+          plain= false;
+#ifdef __WIN__
+          quotation= true;
+#endif
+        }
         /*
           If the equals symbol is encountered, then we need to separately
           process the right side:
@@ -879,13 +933,13 @@ static void copy_orig_argv (char* cmd_str)
               while (m)
               {
                 c = *arg++;
-#ifdef __WIN__
-                if (c == '"' || c == '\\')
-#else
-                if (c == '"' || c == '\\' || c == '$')
-#endif
+                if (IS_REQ_ESCAPING(c))
                 {
+#ifdef __WIN__
+                  *cmd_str++ = c;
+#else
                   *cmd_str++ = '\\';
+#endif
                 }
                 *cmd_str++ = c;
                 m--;
@@ -914,49 +968,20 @@ static void copy_orig_argv (char* cmd_str)
           /* Let's deal with the left side of the expression: */
           while ((c = *arg_scan++) != 0)
           {
-            /*
-              A whitespace or a single quote requires double
-              quotation marks:
-            */
-            if (isspace(c) || c == '\'')
+            if (IS_SPECIAL_V(c))
             {
               quotation= true;
             }
-            /*
-              Double quotation mark or backslash symbol requires backslash
-              prefixing:
-            */
-#ifdef __WIN__
-            else if (c == '"' || c == '\\')
-#else
-            /*
-              The dollar symbol is used to substitute a variable, therefore
-              it also requires escaping:
-            */
-            else if (c == '"' || c == '\\' || c == '$')
-#endif
+            else if (IS_REQ_ESCAPING(c))
             {
               plain= false;
+#ifdef __WIN__
+              quotation= true;
+#endif
             }
           }
           break;
         }
-        /*
-          Double quotation mark or backslash symbol requires backslash
-          prefixing:
-        */
-#ifdef __WIN__
-        else if (c == '"' || c == '\\')
-#else
-        /*
-          The dollar symbol is used to substitute a variable, therefore
-          it also requires escaping:
-        */
-        else if (c == '"' || c == '\\' || c == '$')
-#endif
-        {
-          plain= false;
-        }
       }
       if (n)
       {
@@ -979,13 +1004,13 @@ static void copy_orig_argv (char* cmd_str)
         {
           while ((c = *arg++) != 0)
           {
-#ifdef __WIN__
-            if (c == '"' || c == '\\')
-#else
-            if (c == '"' || c == '\\' || c == '$')
-#endif
+            if (IS_REQ_ESCAPING(c))
             {
+#ifdef __WIN__
+              *cmd_str++ = c;
+#else
               *cmd_str++ = '\\';
+#endif
             }
             *cmd_str++ = c;
           }
@@ -1037,6 +1062,8 @@ static ssize_t sst_prepare_other (const char*  method,
   {
     WSREP_ERROR("sst_prepare_other(): generate_binlog_index_opt_val() failed %d",
                 ret);
+    if (binlog_opt_val) my_free(binlog_opt_val);
+    return ret;
   }
 
   make_wsrep_defaults_file();
@@ -1047,13 +1074,16 @@ static ssize_t sst_prepare_other (const char*  method,
                  WSREP_SST_OPT_ADDR " '%s' "
                  WSREP_SST_OPT_DATA " '%s' "
                  "%s"
-                 WSREP_SST_OPT_PARENT " '%d'"
+                 WSREP_SST_OPT_PARENT " %d "
+                 WSREP_SST_OPT_PROGRESS " %d"
                  "%s"
                  "%s",
                  method, addr_in, mysql_real_data_home,
                  wsrep_defaults_file,
                  (int)getpid(),
+                 0,
                  binlog_opt_val, binlog_index_opt_val);
+
   my_free(binlog_opt_val);
   my_free(binlog_index_opt_val);
 
@@ -1073,7 +1103,7 @@ static ssize_t sst_prepare_other (const char*  method,
     return -env.error();
   }
 
-  if ((ret= sst_append_auth_env(env, sst_auth)))
+  if ((ret= sst_append_env_var(env, WSREP_SST_AUTH_ENV, sst_auth)))
   {
     WSREP_ERROR("sst_prepare_other(): appending auth failed: %d", ret);
     return ret;
@@ -1081,7 +1111,7 @@ static ssize_t sst_prepare_other (const char*  method,
 
   if (data_home_dir)
   {
-    if ((ret= sst_append_data_dir(env, data_home_dir)))
+    if ((ret= sst_append_env_var(env, DATA_HOME_DIR_ENV, data_home_dir)))
     {
       WSREP_ERROR("sst_prepare_other(): appending data "
                   "directory failed: %d", ret);
@@ -1170,6 +1200,19 @@ static ssize_t sst_prepare_mysqldump (const char*  addr_in,
     *addr_out= addr_in;
   }
 
+  pthread_t monitor;
+  ret = mysql_thread_create (key_wsrep_sst_joiner_monitor, &monitor, NULL, wsrep_sst_joiner_monitor_thread, NULL);
+
+  if (ret)
+  {
+    WSREP_ERROR("sst_prepare_other(): mysql_thread_create() failed: %d (%s)",
+                ret, strerror(ret));
+    return -ret;
+  }
+
+  sst_joiner_completed= false;
+  pthread_detach (monitor);
+
   return ret;
 }
 
@@ -1189,35 +1232,46 @@ std::string wsrep_sst_prepare()
   /*
     Figure out SST receive address. Common for all SST methods.
   */
+  wsp::Address* addr_in_parser= NULL;
+
   // Attempt 1: wsrep_sst_receive_address
   if (wsrep_sst_receive_address &&
-      strcmp (wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
+      strcmp(wsrep_sst_receive_address, WSREP_SST_ADDRESS_AUTO))
   {
-    addr_in= wsrep_sst_receive_address;
+    addr_in_parser = new wsp::Address(wsrep_sst_receive_address);
+
+    if (!addr_in_parser->is_valid())
+    {
+      WSREP_ERROR("Could not parse wsrep_sst_receive_address : %s",
+                  wsrep_sst_receive_address);
+      unireg_abort(1);
+    }
   }
-
   //Attempt 2: wsrep_node_address
-  else if (wsrep_node_address && strlen(wsrep_node_address))
+  else if (wsrep_node_address && *wsrep_node_address)
   {
-    wsp::Address addr(wsrep_node_address);
+    addr_in_parser = new wsp::Address(wsrep_node_address);
 
-    if (!addr.is_valid())
+    if (addr_in_parser->is_valid())
+    {
+      // we must not inherit the port number from this address:
+      addr_in_parser->set_port(0);
+    }
+    else
     {
       WSREP_ERROR("Could not parse wsrep_node_address : %s",
                   wsrep_node_address);
       throw wsrep::runtime_error("Failed to prepare for SST. Unrecoverable");
     }
-    memcpy(ip_buf, addr.get_address(), addr.get_address_len());
-    addr_in= ip_buf;
   }
   // Attempt 3: Try to get the IP from the list of available interfaces.
   else
   {
-    ssize_t ret= wsrep_guess_ip (ip_buf, ip_max);
+    ssize_t ret= wsrep_guess_ip(ip_buf, ip_max);
 
     if (ret && ret < ip_max)
     {
-      addr_in= ip_buf;
+      addr_in_parser = new wsp::Address(ip_buf);
     }
     else
     {
@@ -1227,6 +1281,51 @@ std::string wsrep_sst_prepare()
     }
   }
 
+  assert(addr_in_parser);
+
+  size_t len= addr_in_parser->get_address_len();
+  bool is_ipv6= addr_in_parser->is_ipv6();
+  const char* address= addr_in_parser->get_address();
+
+  if (len > (is_ipv6 ? ip_max - 2 : ip_max))
+  {
+    WSREP_ERROR("Address to accept state transfer is too long: '%s'",
+                address);
+    unireg_abort(1);
+  }
+
+  if (is_ipv6)
+  {
+    /* wsrep_sst_*.sh scripts requite ipv6 addreses to be in square breackets */
+    ip_buf[0] = '[';
+    /* the length (len) already includes the null byte: */
+    memcpy(ip_buf + 1, address, len - 1);
+    ip_buf[len] = ']';
+    ip_buf[len + 1] = 0;
+    len += 2;
+  }
+  else
+  {
+    memcpy(ip_buf, address, len);
+  }
+
+  int port= addr_in_parser->get_port();
+  if (port)
+  {
+    size_t space= ip_max - len;
+    ip_buf[len - 1] = ':';
+    int ret= snprintf(ip_buf + len, ip_max - len, "%d", port);
+    if (ret <= 0 || (size_t) ret > space)
+    {
+      WSREP_ERROR("Address to accept state transfer is too long: '%s:%d'",
+                  address, port);
+      unireg_abort(1);
+    }
+  }
+
+  delete addr_in_parser;
+  addr_in = ip_buf;
+
   ssize_t addr_len= -ENOSYS;
   method = wsrep_sst_method;
   if (!strcmp(method, WSREP_SST_MYSQLDUMP))
@@ -1280,7 +1379,7 @@ std::string wsrep_sst_prepare()
 
   const char* method_ptr(ret.data());
   const char* addr_ptr(ret.data() + strlen(method_ptr) + 1);
-  WSREP_INFO ("Prepared SST request: %s|%s", method_ptr, addr_ptr);
+  WSREP_DEBUG("Prepared SST request: %s|%s", method_ptr, addr_ptr);
 
   if (addr_out != addr_in) /* malloc'ed */ free ((char*)addr_out);
 
@@ -1461,20 +1560,18 @@ static int run_sql_command(THD *thd, const char *query)
   }
 
   mysql_parse(thd, thd->query(), thd->query_length(), &ps, FALSE, FALSE);
+
   if (thd->is_error())
   {
     int const err= thd->get_stmt_da()->sql_errno();
-    WSREP_WARN ("Error executing '%s': %d (%s)%s",
-                query, err, thd->get_stmt_da()->message(),
-                err == ER_UNKNOWN_SYSTEM_VARIABLE ?
-                ". Was mysqld built with --with-innodb-disallow-writes ?" : "");
+    WSREP_WARN ("Error executing '%s': %d (%s)",
+                query, err, thd->get_stmt_da()->message());
     thd->clear_error();
     return -1;
   }
   return 0;
 }
 
-
 static int sst_flush_tables(THD* thd)
 {
   WSREP_INFO("Flushing tables for SST...");
@@ -1534,16 +1631,19 @@ static int sst_flush_tables(THD* thd)
   }
   else
   {
+    ha_disable_internal_writes(true);
+
     WSREP_INFO("Tables flushed.");
-    /*
-      Tables have been flushed. Create a file with cluster state ID and
-      wsrep_gtid_domain_id.
-    */
+
+    // Create a file with cluster state ID and wsrep_gtid_domain_id.
     char content[100];
     snprintf(content, sizeof(content), "%s:%lld %d\n", wsrep_cluster_state_uuid,
              (long long)wsrep_locked_seqno, wsrep_gtid_server.domain_id);
     err= sst_create_file(flush_success, content);
 
+    if (err)
+      WSREP_INFO("Creating file for flush_success failed %d",err);
+
     const char base_name[]= "tables_flushed";
     ssize_t const full_len= strlen(mysql_real_data_home) + strlen(base_name)+2;
     char *real_name= (char*) malloc(full_len);
@@ -1578,39 +1678,13 @@ static int sst_flush_tables(THD* thd)
     }
     free(real_name);
     free(tmp_name);
+    if (err)
+      ha_disable_internal_writes(false);
   }
 
   return err;
 }
 
-
-static void sst_disallow_writes (THD* thd, bool yes)
-{
-  char query_str[64]= { 0, };
-  ssize_t const query_max= sizeof(query_str) - 1;
-  CHARSET_INFO *current_charset;
-
-  current_charset= thd->variables.character_set_client;
-
-  if (!is_supported_parser_charset(current_charset))
-  {
-      /* Do not use non-supported parser character sets */
-      WSREP_WARN("Current client character set is non-supported parser character set: %s", current_charset->csname);
-      thd->variables.character_set_client= &my_charset_latin1;
-      WSREP_WARN("For SST temporally setting character set to : %s",
-                 my_charset_latin1.csname);
-  }
-
-  snprintf (query_str, query_max, "SET GLOBAL innodb_disallow_writes=%d",
-            yes ? 1 : 0);
-
-  if (run_sql_command(thd, query_str))
-  {
-    WSREP_ERROR("Failed to disallow InnoDB writes");
-  }
-  thd->variables.character_set_client= current_charset;
-}
-
 static void* sst_donor_thread (void* a)
 {
   sst_thread_arg* arg= (sst_thread_arg*)a;
@@ -1656,20 +1730,32 @@ wait_signal:
       if (!strcasecmp (out, magic_flush))
       {
         err= sst_flush_tables (thd.ptr);
+
         if (!err)
         {
-          sst_disallow_writes (thd.ptr, true);
+          locked= true;
           /*
             Lets also keep statements that modify binary logs (like RESET LOGS,
             RESET MASTER) from proceeding until the files have been transferred
             to the joiner node.
           */
           if (mysql_bin_log.is_open())
-          {
             mysql_mutex_lock(mysql_bin_log.get_log_lock());
-          }
 
-          locked= true;
+          WSREP_INFO("Donor state reached");
+
+#ifdef ENABLED_DEBUG_SYNC
+          DBUG_EXECUTE_IF("sync.wsrep_donor_state",
+          {
+            const char act[]=
+                             "now "
+                             "SIGNAL sync.wsrep_donor_state_reached "
+                             "WAIT_FOR signal.wsrep_donor_state";
+            assert(!debug_sync_set_action(thd.ptr,
+                                          STRING_WITH_LEN(act)));
+          };);
+#endif
+
           goto wait_signal;
         }
       }
@@ -1677,14 +1763,11 @@ wait_signal:
       {
         if (locked)
         {
-          if (mysql_bin_log.is_open())
-          {
-            mysql_mutex_assert_owner(mysql_bin_log.get_log_lock());
-            mysql_mutex_unlock(mysql_bin_log.get_log_lock());
-          }
-          sst_disallow_writes (thd.ptr, false);
-          thd.ptr->global_read_lock.unlock_global_read_lock(thd.ptr);
           locked= false;
+          ha_disable_internal_writes(false);
+          if (mysql_bin_log.is_open())
+            mysql_mutex_unlock(mysql_bin_log.get_log_lock());
+          thd.ptr->global_read_lock.unlock_global_read_lock(thd.ptr);
         }
         err=  0;
         goto wait_signal;
@@ -1715,12 +1798,12 @@ wait_signal:
 
   if (locked) // don't forget to unlock server before return
   {
+    ha_disable_internal_writes(false);
     if (mysql_bin_log.is_open())
     {
       mysql_mutex_assert_owner(mysql_bin_log.get_log_lock());
       mysql_mutex_unlock(mysql_bin_log.get_log_lock());
     }
-    sst_disallow_writes (thd.ptr, false);
     thd.ptr->global_read_lock.unlock_global_read_lock(thd.ptr);
   }
 
@@ -1767,6 +1850,8 @@ static int sst_donate_other (const char*        method,
   {
     WSREP_ERROR("sst_prepare_other(): generate_binlog_index_opt_val() failed %d",
                 ret);
+    if (binlog_opt_val) my_free(binlog_opt_val);
+    return ret;
   }
 
   make_wsrep_defaults_file();
@@ -1777,15 +1862,19 @@ static int sst_donate_other (const char*        method,
                  "wsrep_sst_%s "
                  WSREP_SST_OPT_ROLE " 'donor' "
                  WSREP_SST_OPT_ADDR " '%s' "
+                 WSREP_SST_OPT_LPORT " %u "
                  WSREP_SST_OPT_SOCKET " '%s' "
+                 WSREP_SST_OPT_PROGRESS " %d "
                  WSREP_SST_OPT_DATA " '%s' "
                  "%s"
                  WSREP_SST_OPT_GTID " '%s:%lld' "
-                 WSREP_SST_OPT_GTID_DOMAIN_ID " '%d'"
+                 WSREP_SST_OPT_GTID_DOMAIN_ID " %d"
                  "%s"
                  "%s"
                  "%s",
-                 method, addr, mysqld_unix_port, mysql_real_data_home,
+                 method, addr, mysqld_port, mysqld_unix_port,
+                 0,
+                 mysql_real_data_home,
                  wsrep_defaults_file,
                  uuid_oss.str().c_str(), gtid.seqno().get(), wsrep_gtid_server.domain_id,
                  binlog_opt_val, binlog_index_opt_val,
@@ -1858,11 +1947,6 @@ int wsrep_sst_donate(const std::string& msg,
                      const wsrep::gtid& current_gtid,
                      const bool         bypass)
 {
-  /* This will be reset when sync callback is called.
-   * Should we set wsrep_ready to FALSE here too? */
-
-  wsrep_config_state->set(wsrep::server_state::s_donor);
-
   const char* method= msg.data();
   size_t method_len= strlen (method);
 
@@ -1874,7 +1958,21 @@ int wsrep_sst_donate(const std::string& msg,
 
   const char* data= method + method_len + 1;
 
-  if (check_request_str(data, address_char))
+  /* check for auth@addr separator */
+  const char* addr= strrchr(data, '@');
+  wsp::string remote_auth;
+  if (addr)
+  {
+    remote_auth.set(strndup(data, addr - data));
+    addr++;
+  }
+  else
+  {
+    // no auth part
+    addr= data;
+  }
+
+  if (check_request_str(addr, address_char))
   {
     WSREP_ERROR("Bad SST address string. SST canceled.");
     return WSREP_CB_FAILURE;
@@ -1888,15 +1986,25 @@ int wsrep_sst_donate(const std::string& msg,
   }
 
   int ret;
-  if ((ret= sst_append_auth_env(env, sst_auth_real)))
+  if ((ret= sst_append_env_var(env, WSREP_SST_AUTH_ENV, sst_auth_real)))
   {
     WSREP_ERROR("wsrep_sst_donate_cb(): appending auth env failed: %d", ret);
     return WSREP_CB_FAILURE;
   }
 
+  if (remote_auth())
+  {
+    if ((ret= sst_append_env_var(env, WSREP_SST_REMOTE_AUTH_ENV,remote_auth())))
+    {
+      WSREP_ERROR("wsrep_sst_donate_cb(): appending remote auth env failed: "
+                  "%d", ret);
+      return WSREP_CB_FAILURE;
+    }
+  }
+
   if (data_home_dir)
   {
-    if ((ret= sst_append_data_dir(env, data_home_dir)))
+    if ((ret= sst_append_env_var(env, DATA_HOME_DIR_ENV, data_home_dir)))
     {
       WSREP_ERROR("wsrep_sst_donate_cb(): appending data "
                   "directory failed: %d", ret);
@@ -1918,11 +2026,11 @@ int wsrep_sst_donate(const std::string& msg,
 
   if (!strcmp (WSREP_SST_MYSQLDUMP, method))
   {
-    ret= sst_donate_mysqldump(data, current_gtid, bypass, env());
+    ret= sst_donate_mysqldump(addr, current_gtid, bypass, env());
   }
   else
   {
-    ret= sst_donate_other(method, data, current_gtid, bypass, env());
+    ret= sst_donate_other(method, addr, current_gtid, bypass, env());
   }
 
   return (ret >= 0 ? 0 : 1);
