@@ -1,6 +1,6 @@
 /* Copyright (c) 2002-2007 MySQL AB & tommy@valley.ne.jp
    Copyright (c) 2002, 2014, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2014, SkySQL Ab.
+   Copyright (c) 2009, 2020, MariaDB Corporation.
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -15,7 +15,7 @@
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301, USA */
+   MA 02110-1335  USA */
 
 /* This file is for binary pseudo charset, created by bar@mysql.com */
 
@@ -23,6 +23,8 @@
 #include "strings_def.h"
 #include <m_ctype.h>
 
+const char charset_name_binary[]= "binary";
+
 static const uchar ctype_bin[]=
 {
   0,
@@ -125,6 +127,18 @@ static int my_strnncollsp_binary(CHARSET_INFO * cs __attribute__((unused)),
 }
 
 
+static int my_strnncollsp_nchars_binary(CHARSET_INFO * cs __attribute__((unused)),
+                                        const uchar *s, size_t slen,
+                                        const uchar *t, size_t tlen,
+                                        size_t nchars,
+                                        uint flags)
+{
+  set_if_smaller(slen, nchars);
+  set_if_smaller(tlen, nchars);
+  return my_strnncoll_binary(cs, s, slen, t, tlen, 0);
+}
+
+
 static int my_strnncoll_8bit_bin(CHARSET_INFO * cs __attribute__((unused)),
                                  const uchar *s, size_t slen,
                                  const uchar *t, size_t tlen,
@@ -199,6 +213,18 @@ static int my_strnncollsp_8bit_bin(CHARSET_INFO * cs __attribute__((unused)),
 }
 
 
+static int my_strnncollsp_nchars_8bit_bin(CHARSET_INFO * cs,
+                                          const uchar *a, size_t a_length,
+                                          const uchar *b, size_t b_length,
+                                          size_t nchars,
+                                          uint flags)
+{
+  set_if_smaller(a_length, nchars);
+  set_if_smaller(b_length, nchars);
+  return my_strnncollsp_8bit_bin(cs, a, a_length, b, b_length);
+}
+
+
 static int my_strnncollsp_8bit_nopad_bin(CHARSET_INFO * cs
                                          __attribute__((unused)),
                                          const uchar *a, size_t a_length,
@@ -219,7 +245,7 @@ static size_t my_case_str_bin(CHARSET_INFO *cs __attribute__((unused)),
 
 static size_t my_case_bin(CHARSET_INFO *cs __attribute__((unused)),
                           const char *src, size_t srclen,
-                          char *dst, size_t dstlen)
+                          char *dst, size_t dstlen __attribute__((unused)))
 {
   DBUG_ASSERT(srclen <= dstlen);
   memcpy(dst, src, srclen);
@@ -399,7 +425,7 @@ my_strnxfrm_8bit_bin(CHARSET_INFO *cs,
 {
   set_if_smaller(srclen, dstlen);
   set_if_smaller(srclen, nweights);
-  if (dst != src)
+  if (srclen && dst != src)
     memcpy(dst, src, srclen);
   return my_strxfrm_pad_desc_and_reverse(cs, dst, dst + srclen, dst + dstlen,
                                          (uint)(nweights - srclen), flags, 0);
@@ -487,6 +513,7 @@ MY_COLLATION_HANDLER my_collation_8bit_bin_handler =
   my_coll_init_8bit_bin,
   my_strnncoll_8bit_bin,
   my_strnncollsp_8bit_bin,
+  my_strnncollsp_nchars_8bit_bin,
   my_strnxfrm_8bit_bin,
   my_strnxfrmlen_simple,
   my_like_range_simple,
@@ -503,6 +530,7 @@ MY_COLLATION_HANDLER my_collation_8bit_nopad_bin_handler =
   my_coll_init_8bit_bin,
   my_strnncoll_8bit_bin,
   my_strnncollsp_8bit_nopad_bin,
+  my_strnncollsp_nchars_8bit_bin,
   my_strnxfrm_8bit_nopad_bin,
   my_strnxfrmlen_simple,
   my_like_range_simple,
@@ -519,6 +547,7 @@ static MY_COLLATION_HANDLER my_collation_binary_handler =
   NULL,			/* init */
   my_strnncoll_binary,
   my_strnncollsp_binary,
+  my_strnncollsp_nchars_binary,
   my_strnxfrm_8bit_bin,
   my_strnxfrmlen_simple,
   my_like_range_simple,
@@ -560,6 +589,7 @@ static MY_CHARSET_HANDLER my_charset_handler=
   my_well_formed_char_length_8bit,
   my_copy_8bit,
   my_wc_mb_bin,
+  my_wc_to_printable_generic
 };
 
 
@@ -567,7 +597,7 @@ struct charset_info_st my_charset_bin =
 {
     63,0,0,			/* number        */
     MY_CS_COMPILED|MY_CS_BINSORT|MY_CS_PRIMARY|MY_CS_NOPAD,/* state */
-    "binary",			/* cs name    */
+    charset_name_binary,	/* cs name    */
     "binary",			/* name          */
     "",				/* comment       */
     NULL,			/* tailoring     */
